import gym
from gym import logger
import numpy as np
from gym.envs.classic_control import rendering
from gym import spaces

class SwitchEnv(gym.Env):
	"""
	Description:
		Two Agents in a 2D-grid partially observable environment. The task is for both agents to reach the switch at the opposite side of the map.
		In the middle theres a corridor, to narrow for both agents to pass.

	Source:
		The environment appeared in the paper Value-Decomposition Networks FOr Cooperative Multi-Agent Learning. (https://arxiv.org/pdf/1706.05296.pdf)

	Observation:
		Type: np.array of shape (2, view, view) consisting of the partial observations for the two agents:
		Num		Obs
		0		Free space
		1		Wall
		2		Agent 0
		3 		Agent 1
		4		Switch for Agent 0
		5		Switch for Agent 1

	Actions:
		Type: np.array of shape (2). For each entry:
		Num    Action
		0	   Move forward
		1	   Turn left
		2      Turn right

	Reward:
		If an agent reaches its switch for the first time the reward is 1.

	Starting State:
		 Both agents start on opposing sides of the map with their respective observation.

	Episode Termination:
		 After both agents have found their respective switch.
	"""

	metadata = {"render.modes": ["human"]}

	def __init__(self, height: int = 7, width: int = 30, view: int = 5, flatten_obs: bool = False):
		self.pixel_map = self._create_corridor_map(height, width, view - 1)
		if view % 2 == 0:
			raise ValueError("View must be odd integer value such that agent can be centered.")
		self.view = view
		self.width = width
		self.height = height
		self.flatten = flatten_obs
		self.agents_pos = [[view-1+height//2, view-1, 1], [view-1+height//2, width+(view-1)-1, 3]]
		self.switch_pos = [[view-1+height//2, 11*width//12+(view-1)-1], [view-1+height//2, width//12+view]]
		self.switch_state = [0, 0]
		self.received_reward = False
		self.done = 0
		self.viewer = None

		obs_space = spaces.Box(low=0, high=5, shape=(self.view+1 if self.view % 2 == 0 else self.view, self.view))
		self.observation_space = spaces.Tuple((obs_space, obs_space))

		self.action_space = spaces.Tuple((spaces.Discrete(3), spaces.Discrete(3)))

	def step(self, act: np.array):
		
		if self.done:
			logger.warn(
				"You are calling 'step()' even though this "
				"environment has already returned done = True. You "
				"should always call 'reset()' once you receive 'done = "
				"True' -- any further steps are undefined behavior."
			)

		# act
		self._act_agent(0, act[0])				
		self._act_agent(1, act[1])

		# get observations
		obs = np.stack((self._get_observation(0), self._get_observation(1)))

		# both agents reach goal simultaneously
		if sum(self.switch_state) == 2 and not self.received_reward:
			rew = 2
			self.done = 1
		# both agents reached goal
		elif sum(self.switch_state) == 2:
			rew = 1
			self.done = 1
		# one agent reached goal
		elif sum(self.switch_state) == 1 and not self.received_reward:
			rew = 1
			self.done = 0
			self.received_reward = True
		# no agent reached goal
		else:
			rew = 0
			self.done = 0

		if self.flatten:
			obs = obs.reshape((2, self.view**2))

		return obs, rew, self.done, {}
	
	def reset(self) -> np.array:

		self.pixel_map[self.agents_pos[0][0], self.agents_pos[0][1]] = 0
		self.pixel_map[self.agents_pos[1][0], self.agents_pos[1][1]] = 0
		self.pixel_map[self.switch_pos[0][0], self.switch_pos[0][1]] = 4
		self.pixel_map[self.switch_pos[1][0], self.switch_pos[1][1]] = 5

		self.agents_pos = [[self.view-1+self.height//2, self.view-1, 1], [self.view-1+self.height//2, self.width+(self.view-1)-1, 3]]
		self.switch_pos = [[self.view-1+self.height//2, 11*self.width//12+(self.view-1)-1], [self.view-1+self.height//2, self.width//12+self.view]]
		
		self.pixel_map[self.agents_pos[0][0], self.agents_pos[0][1]] = 2
		self.pixel_map[self.agents_pos[1][0], self.agents_pos[1][1]] = 3

		self.switch_state = [0, 0]
		obs = np.stack((self._get_observation(0), self._get_observation(1)))

		if self.flatten:
			obs = obs.reshape((2, self.view**2))

		self.done = 0
		
		return obs

	def render(self, mode: str = "human"):
		
		if self.viewer == None:
			self.viewer = rendering.Viewer((100*self.pixel_map.shape[1])//4, (100*self.pixel_map.shape[0])//4)
			self.viewer.set_bounds(0, self.pixel_map.shape[1], 0, self.pixel_map.shape[0])
			for i in range(self.pixel_map.shape[0]):
				for j in range(self.pixel_map.shape[1]):
					if self.pixel_map[i][j] == 1:
						wall = rendering.FilledPolygon([(j, i), (j, i+1), (j+1, i+1), (j+1, i)])
						self.viewer.add_geom(wall)

		for i in range(self.pixel_map.shape[0]):
			for j in range(self.pixel_map.shape[1]):
				
				if self.pixel_map[i][j] != 0:
					x = j
					y = self.pixel_map.shape[0] - i - 1

				if self.pixel_map[i][j] == 2:
					agent, view = self._get_rendered_agent(0, x, y, self.agents_pos[0][2])
					self.viewer.add_onetime(agent)
					self.viewer.add_onetime(view)

				elif self.pixel_map[i][j] == 3:
					agent, view = self._get_rendered_agent(1, x, y, self.agents_pos[1][2])
					self.viewer.add_onetime(agent)
					self.viewer.add_onetime(view)

				elif self.pixel_map[i][j] == 4:
					switch = self._get_rendered_switch(0, x, y)
					self.viewer.add_onetime(switch)

				elif self.pixel_map[i][j] == 5:
					switch = self._get_rendered_switch(1, x, y)
					self.viewer.add_onetime(switch)

		self.viewer.render(return_rgb_array=mode == "rgb_array")


	def close(self):
		if self.viewer:
			self.viewer.close()
			self.viewer = None	


	def _get_observation(self, agent_id: int) -> np.array:
		
		y = self.agents_pos[agent_id][0]
		x = self.agents_pos[agent_id][1]
		rot = self.agents_pos[agent_id][2]

		if rot == 0:
			obs = self.pixel_map[y-self.view+1:y+1, x-self.view//2:x+self.view//2+1]
		elif rot == 1:
			obs = self.pixel_map[y-self.view//2:y+self.view//2+1, x:x+self.view]
		elif rot == 2:
			obs = self.pixel_map[y:y+self.view, x-self.view//2:x+self.view//2+1]
		elif rot == 3:
			obs = self.pixel_map[y-self.view//2:y+self.view//2+1, x-self.view+1:x+1]
		obs = np.rot90(m=obs, k=rot)
		return obs


	def _rotate_agent(self, agent_id: int, right: bool):
		
		self.agents_pos[agent_id][2] += (1 if right else -1)
		self.agents_pos[agent_id][2] %= 4


	def _move_agent(self, agent_id: int):
		
		y = self.agents_pos[agent_id][0]
		x = self.agents_pos[agent_id][1]
		rot = self.agents_pos[agent_id][2]
		
		if rot == 0 and self.pixel_map[y-1, x] not in [1, 2, 3]: 
			self.agents_pos[agent_id][0] -= 1
			self.pixel_map[y, x] = 0
			self.pixel_map[y-1, x] = agent_id + 2
		elif rot == 1 and self.pixel_map[y, x+1] not in [1, 2, 3]:
			self.agents_pos[agent_id][1] += 1
			self.pixel_map[y, x] = 0
			self.pixel_map[y, x+1] = agent_id + 2
		elif rot == 2 and self.pixel_map[y+1, x] not in [1, 2, 3]:
			self.agents_pos[agent_id][0] += 1
			self.pixel_map[y, x] = 0
			self.pixel_map[y+1, x] = agent_id + 2
		elif rot == 3 and self.pixel_map[y, x-1] not in [1, 2, 3]:
			self.agents_pos[agent_id][1] -= 1
			self.pixel_map[y, x] = 0
			self.pixel_map[y, x-1] = agent_id + 2
	
		for switch_id in range(2):
			visible = True
			for agent_id in range(2):	
				if self.agents_pos[agent_id][0] == self.switch_pos[switch_id][0] and self.agents_pos[agent_id][1] == self.switch_pos[switch_id][1]:
					visible = False
					if agent_id == switch_id:
						self.switch_state[switch_id] = 1
			if visible:
				self.pixel_map[self.switch_pos[switch_id][0], self.switch_pos[switch_id][1]] = switch_id + 4

	
	def _act_agent(self, agent_id: int, act: int):
		
		if act == 0:
			self._move_agent(agent_id)
		elif act == 1 or act == 2:
			self._rotate_agent(agent_id, right=act-1)


	def _get_rendered_agent(self, agent_id: int, x: int, y: int, rot: int) -> rendering.Geom:
		
		if rot == 0:
			agent = rendering.FilledPolygon([(x + 1/2, y + 5/6), (x + 4/5, y + 1/6), (x + 1/5, y + 1/6)])
			view = rendering.PolyLine([(x-self.view//2, y+self.view), (x+self.view//2+1, y+self.view), (x+self.view//2+1, y), (x-self.view//2, y)],close=True)
		elif rot == 1:
			agent = rendering.FilledPolygon([(x + 5/6, y + 1/2), (x + 1/6, y + 4/5), (x + 1/6, y + 1/5)])
			view = rendering.PolyLine([(x, y+self.view//2+1), (x+self.view, y+self.view//2+1), (x+self.view, y-self.view//2), (x, y-self.view//2)],close=True)
		elif rot == 2:
			agent = rendering.FilledPolygon([(x + 1/2, y + 1/6), (x + 4/5, y + 5/6), (x + 1/5, y + 5/6)])
			view = rendering.PolyLine([(x-self.view//2, y-self.view+1), (x+self.view//2+1, y-self.view+1), (x+self.view//2+1, y+1), (x-self.view//2, y+1)],close=True)
		elif rot == 3:
			agent = rendering.FilledPolygon([(x + 1/6, y + 1/2), (x + 5/6, y + 4/5), (x + 5/6, y + 1/5)])
			view = rendering.PolyLine([(x+1, y+self.view//2+1), (x-self.view+1, y+self.view//2+1), (x-self.view+1, y-self.view//2), (x+1, y-self.view//2)],close=True)
		agent.set_color(*((1, 0, 0) if agent_id else (0, 0, 1)))
		view.set_color(*((1, 0, 0) if agent_id else (0, 0, 1)))
		
		return agent, view


	def _get_rendered_switch(self, switch_id: int, x: int, y: int) -> rendering.Geom:
		switch = rendering.FilledPolygon([(x + 1/2, y + 1/4), (x + 3/4, y + 1/2), (x + 1/2, y + 3/4), (x + 1/4, y + 1/2)])
		switch.set_color(*((1, 0, 0) if switch_id else (0, 0, 1)))
		return switch

	def _create_corridor_map(self, height: int, width: int, padding: int) -> np.array:
		pixel_map = []
		room_width = (width + 1)//4
		corridor_length = width - 2*room_width
		corridor_position = height//2

		for h in range(height):
			pixel_map.append([])
			for w in range(width):
				if w >= room_width and w < room_width + corridor_length and h != corridor_position:
					tile = 1
				else:
					tile = 0
				pixel_map[h].append(tile)

		return np.pad(np.array(pixel_map), padding, 'constant', constant_values=(1))



import gym
from gym import logger, spaces
import numpy as np
import pygame
from pygame import gfxdraw

class NavigationEnv(gym.Env):
	"""
	Description:
		N Agents cooperatively try to cover a set of N landmarks while avoiding colliding with other agents.
	Source:
		The environment appeared in the MADDPG paper (https://arxiv.org/pdf/1706.02275.pdf).
	Observation:
		Type: Box(n_agents, n_agents*4)
        Num    						 		    Observation               		  Min                             Max
        [i < n_agents, 2*j    <  n_agents*2]    Relative xPos agent to agent      -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size
        [i < n_agents, 2*j+1  <  n_agents*2]    Relative yPos agent to agent      -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size
        [i < n_agents, n_agents*2 <=    2*j]    Relative xPos agent to landmark   -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size
        [i < n_agents, n_agents*2 <=  2*j+1]    Relative xPos agent to landmark   -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size

	Actions:
		Type: Box(n_agents, 2)
        Num    						 		 	Observation               		  Min                             Max
		[i < n_agents, 0]						Angle of movement				  -pi							  pi
		[i < n_agents, 1]						Distance of movement			  0.0							  1.0

	Reward:
		 The agents are punished for uncovered landmarks and for colliding. The trade off between both can be adjusted with the parameter tau:
		 reward = - (tau * n_uncovered + (1-tau) * n_colliding)
 
				
	Starting State:
		 Agents start at random position in (2.5*N)x(2.5*N) units big environment one agents diameter is 1 unit.

	Episode Termination:
		 Either after max_steps or after keeping all landmarks covered for hold_steps.
	"""

	metadata = {"render.modes": ["human"]}

	def __init__(self, n_agents, world_size: float = 2.5, tau: float = 0.5, max_steps: int = np.inf,
                 hold_steps: int = 10, sparse: bool = False, init_agent_pos: np.array = None, init_landmark_pos: np.array = None):
		
		self.border = world_size*n_agents
		self.n_agents = n_agents
		self.tau = tau
		self.sparse = sparse

		self.init_agent_pos = None if init_agent_pos is None else init_agent_pos
		self.init_landmark_pos = None if init_landmark_pos is None else init_landmark_pos
		
		self.low_state = np.array(
        	[[-self.border for j in range(n_agents*4 - 2)] for i in range(n_agents)], dtype=np.float32
        )

		self.high_state = np.array(
        	[[self.border for j in range(n_agents*4 - 2)] for i in range(n_agents)], dtype=np.float32
        )

		self.action_space = spaces.Box(
			low=np.array([[-np.pi, 0.0]]*n_agents, dtype=np.float32), high=np.array([[np.pi, 1.0]]*n_agents, dtype=np.float32), dtype=np.float32
        )
		
		self.observation_space = spaces.Box(
			low=self.low_state, high=self.high_state, dtype=np.float32
        )

		
		# min reward for all situations -> all landmarks in one corner all agents in the corner diagonally across
		# -> maximum distance from landmarks + maximal overlapping
		self.agents = np.full(shape=(n_agents, 2), fill_value=self.border - 0.5)
		self.landmarks = np.full(shape=(n_agents, 2), fill_value=0.25)
		_, dist = self._get_observations()
		min_rew, _ = self._calculate_reward(dist)

		self.reward_space = spaces.Box(
			low=min_rew, high=0, shape=(1,)
		)
	
		if self.init_agent_pos is None:
			self.agents = np.random.rand(self.n_agents, 2)*(self.border - 1) + 0.5
		else:
			if self.init_agent_pos.shape != (self.n_agents, 2):
				raise ValueError(f"Agent position shape incorrect. Should be: {(self.n_agents, 2)}. Is {self.init_agent_pos.shape}")
			self.agents = self.init_agent_pos.copy()

		if self.init_landmark_pos is None:
			self.landmarks = np.random.rand(self.n_agents, 2)*(self.border - 0.5) + 0.25
		else:
			if self.init_landmark_pos.shape != (self.n_agents, 2):
				raise ValueError(f"Landmark position shape incorrect. Should be: {(self.n_agents, 2)}. Is {self.init_landmark_pos.shape}")
			self.landmarks = self.init_landmark_pos.copy()

		self.done = False
		self.step_counter = 0
		self.max_steps = max_steps
		
		self.hold_step_counter = 0
		self.hold_steps = hold_steps

		self.surface = None
		self.screen = None

	def step(self, act: np.array):
		
		if self.done:
			logger.warn(
				"You are calling 'step()' even though this "
				"environment has already returned done = True. You "
				"should always call 'reset()' once you receive 'done = "
				"True' -- any further steps are undefined behavior."
			)

		self._check_action(act)
		trans = [[r*np.cos(phi), r*np.sin(phi)] for phi, r in act]
		self.agents += trans
		self.agents = np.clip(self.agents, 0.5, self.border-0.5)
		obs, dist = self._get_observations()
		
		rew, n_uncovered = self._calculate_reward(dist)

		self.step_counter += 1

		if n_uncovered == 0:
			self.hold_step_counter += 1
		
		else:
			self.hold_step_counter = 0

		if self.step_counter >= self.max_steps or self.hold_step_counter >= self.hold_steps:
			self.done = True
		
		return obs, rew, self.done, {}


	def reset(self) -> np.array:
		
		if self.init_agent_pos is None:
			self.agents = np.random.rand(self.n_agents, 2)*(self.border - 1) + 0.5
		else:
			self.agents = self.init_agent_pos.copy()

		if self.init_landmark_pos is None:
			self.landmarks = np.random.rand(self.n_agents, 2)*(self.border - 0.5) + 0.25
		else:
			self.landmarks = self.init_landmark_pos.copy()

		self.done = False
		self.step_counter = 0
		self.hold_step_counter = 0
		obs, _ = self._get_observations()
		
		return obs


	def _calculate_reward(self, dist):
	
		# number of agent collistions (all distances between two agents smaller than agent diameter. Divided by two due to every distance twice in matrix) 
		n_collisions = (np.count_nonzero(dist[:, :self.n_agents - 1] < 1))//2
		
		# number of uncovered landmarks
		n_uncovered = np.count_nonzero(np.all(dist[:, self.n_agents - 1:] > 0.25, axis=0))
		
		# sum of distances of closest agent for each landmark
		min_dist = np.sum(np.min(dist[:, self.n_agents - 1:], axis=0))
		
		if self.sparse:
			rew = - (self.tau * n_uncovered + (1-self.tau) * n_collisions)
	
		else:
			rew = - (self.tau * min_dist + (1-self.tau) * n_collisions)

		return rew, n_uncovered
	

	def _check_action(self, action):
		
		if action.shape != (self.n_agents, 2):
			raise ValueError(f"Action has wrong shape. Expected ({self.n_agents}, 2), got {action.shape}.")
				

	def _get_observations(self) -> np.array:
		# n_agents x (n_agents - 1 + n_landmarks)*2
		observations = np.zeros((self.n_agents, (2*self.n_agents - 1)*2))
		# n_agents x (n_agents - 1 + n_landmarks)
		dist = np.zeros((self.n_agents, 2*self.n_agents - 1))

		for i in range(self.n_agents):
			for j in range(self.n_agents):
				
				# agent-landmark x, y distances
				index = (self.n_agents-1)*2 + 2*j
				observations[i, index:index + 2] = self.agents[i] - self.landmarks[j]
				
				# agent-landmark direct distances
				dist[i, self.n_agents-1 + j] = np.linalg.norm(self.agents[i] - self.landmarks[j], ord=1)
				
				# agent-agent not with itself
				if i == j:
					continue
				
				k = j if j < i else j - 1

				# agent-agent x, y distances
				observations[i, 2*k:2*k+2] = self.agents[i] - self.agents[j]
				
				# agent-landmark direct distances
				dist[i, k] = np.linalg.norm(self.agents[i] - self.agents[j], ord=1)			
		
		return observations, dist

	# Render environment
	def render(self, mode="human"):
		
		if self.surface == None:
			length = int(self._world_to_render(self.border))
			self.screen = pygame.display.set_mode((length, length))
			self.surface = pygame.Surface((length, length))
		
		self.surface.fill((255, 255, 255))

		for landmark in self.landmarks:
			gfxdraw.filled_circle(self.surface, self._world_to_render(landmark[0]), self._world_to_render(landmark[1]), self._world_to_render(0.25), (255, 0, 0))

		for agent in self.agents:
			gfxdraw.filled_circle(self.surface, self._world_to_render(agent[0]), self._world_to_render(agent[1]), self._world_to_render(0.5), (0, 255, 0))
		
		self.screen.blit(self.surface, (0, 0))
		pygame.display.flip()
	
	def _world_to_render(self, coordinates):
		if isinstance(coordinates, int) or isinstance(coordinates, float):
			return int((100*coordinates)//4)
		elif isinstance(coordinates, list):
			return [int((c*100)//4) for c in coordinates]
		elif isinstance(coordinates, tuple):
			return (int((c*100)//4) for c in coordinates)

	def close(self):
		if self.screen:
			self.surface = None
			self.screen = None
			pygame.quit()

	def get_obs_dim(self):
		return self.n_agents*4 - 2

	def get_act_dim(self):
		return 2



if __name__ == '__main__':
	
	from marlenvs.wrappers import NormalizeActWrapper, NormalizeObsWrapper
	import time

	env = NavigationEnv(n_agents = 3, max_steps = 100, tau=1.0, world_size=5)
	env = NormalizeActWrapper(env)
	env = NormalizeObsWrapper(env)
	env.reset()
	env.render()
	while True:
		obs, rew, _, _ =  env.step(np.array([[0, 0], [0, 0], [0, 0]]))
		env.render()
		time.sleep(0.03)



import gym

class Mapper:

	def __init__(self, low_in, high_in, low_out, high_out):
		
		self.scale = (high_out - low_out)/(high_in - low_in)
		self.low_out = low_out
		self.low_in = low_in

	def map(self, value):

		return self.low_out + self.scale * (value - self.low_in)


class NormalizeObsWrapper(gym.ObservationWrapper):

	def __init__(self, env, high: float = 1.0, low: float = 0.0):
		super().__init__(env)
		self.mapper = Mapper(low_in = env.observation_space.low, high_in = env.observation_space.high, low_out = low, high_out = high)
	
	def observation(self, obs):
		return self.mapper.map(obs)

class NormalizeActWrapper(gym.ActionWrapper):

	def __init__(self, env, high: float = 1.0, low: float = 0.0):
		super().__init__(env)
		self.mapper = Mapper(low_in = low, high_in = high, low_out = env.action_space.low, high_out = env.action_space.high)

	def action(self, act):
		return self.mapper.map(act)

class NormalizeRewWrapper(gym.RewardWrapper):

	"""
	random policy zero calculates average reward of random policy and sets it such that 0 is at that value.
	"""
	def __init__(self, env, high: float = 1.0, low: float = 0.0, random_policy_zero: bool = False):
		super().__init__(env)
		self.mapper = Mapper(low_in = env.reward_space.low[0], high_in = env.reward_space.high[0], low_out = low, high_out = high)

	def reward(self, rew):
		return self.mapper.map(rew)


def get_avg_random_reward(env):
	avg_reward = 0
	for episode in range(1000):
		env.reset()
		for steps in range(100):
			_, rew, _, _ = env.step(env.action_space.sample())
			avg_reward += rew
	return avg_reward/100000



from gym.envs.registration import register

register(id='TwoStep-v0', entry_point='marlenvs.two_step_env:TwoStepEnv')
register(id='TwoStepCont-v0', entry_point='marlenvs.two_step_cont_env:TwoStepContEnv')
register(id='Switch-v0', entry_point='marlenvs.switch_env:SwitchEnv')
register(id='Navigation-v0', entry_point='marlenvs.navigation_env:NavigationEnv')



from setuptools import setup, find_packages

setup(
		name='marlenvs',
		version='1.0.10',
		author='Lukas König',
		author_email='lukasmkoenig@gmx.net',
        packages=find_packages(),
		install_requires=['gym', 'numpy', 'wheel', 'pyglet', 'pygame'],
		decription="Multi-Agent Reinforcement Learning environments for gym."
	)



import gym
from gym import logger, spaces
import numpy as np
import matplotlib.pyplot as plt


class TwoStepContEnv(gym.Env):
	"""
	Description:
		Continous version of two step environment wit variable amount of agents.
	Source:
		Own idea based on discrete environment which appeared in the QMIX paper.
	Observation:
		Type: Box(n_agents, 1)
		Num    Observation
		gives 1 for being first agent, 0 else	

	Actions:
		Type: Box(n_agents, 1)
		Num    Action         		  Min  Max
		 i	   Action agent i		   0    1

	Reward:
		The agents get a reward only in the final state based on this formula:
			reward avg(all_final_actions)**2 * agent1_first_action + (1 - agent1_first_action) * self.intersection
				
	Episode Termination:
		 After two rounds
	"""

	metadata = {"render.modes": ["human"]}

	def __init__(self, n_agents: int  = 2, intersection: int = 0.5):
		
		self.game_state = 0
		self.intersection = intersection
		self.n_agents = n_agents


		self.action_space = spaces.Box(
			low = 0, high = 1, shape = (self.n_agents, 1)
		)
		
		self.observation_space = spaces.Box(
			low = 0, high = 1, shape = (self.n_agents, 1)
		)

		self.reward_space = spaces.Box(
			low = 0, high = 1, shape = (1,)
		)
		
		# actions
		self.weight = None
		self.actions = None
		
		# render stuff
		self.fig = None
		self.ax = None
		self.ln1 = None
		self.ln2 = None
		self.ln3 = None

	def step(self, act):
		
		if self.game_state == 0:
			self.weight = 1.0 - act[0][0]
			rew = 0.0
			done = 0

		elif self.game_state == 1:
			self.actions = np.squeeze(act)
			rew = self._reward(np.mean(self.actions))
			done = 1

		else:
			logger.warn(
				"You are calling 'step()' even though this "
				"environment has already returned done = True. You "
				"should always call 'reset()' once you receive 'done = "
				"True' -- any further steps are undefined behavior."
			)

		obs = np.array([[1]] + [[0]]*(self.n_agents - 1))
		self.game_state += 1
		
		return obs, rew, done, {}

	def reset(self):
		self.game_state = 0
		self.weight = None
		self.actions = None

		obs = np.array([[1]] + [[0]]*(self.n_agents - 1))
		return obs

	def _reward(self, t):
		return t**2 * self.weight + (1 - self.weight) * self.intersection

	def render(self, mode="human"):
		
		resolution = 1000
		t = [v/resolution for v in range(resolution)]
		w_0 = [self.intersection for v in t]
		w_1 = [v**2 * 1 for v in t]
	
		if self.fig is None:
			plt.ion()
			self.fig, self.ax = plt.subplots(1)
			self.ax.fill_between(t, w_0, w_1, facecolor='blue', alpha=0.5, zorder=1)
			self.ax.set_xlabel('average action value')
			self.ax.set_ylabel('reward')

	
		if self.game_state > 0:
			
			if self.ln1 is not None:
				self.ln1.remove()
				self.ln1 = None
			if self.ln2 is not None:
				self.ln2.remove()
				self.ln2 = None
			if self.ln3 is not None:
				self.ln3.remove()
				self.ln3 = None
	
			f = [self._reward(v) for v in t]
			self.ln1, = self.ax.plot(t, f, color='red', zorder=2)
		
		if self.game_state > 1:
			r = [self._reward(v) for v in self.actions]
			avg = self._reward(np.mean(self.actions))
			self.ln2 = self.ax.scatter(self.actions, r, color='red', zorder=3)
			self.ln3 = self.ax.scatter(np.mean(self.actions), avg, color='yellow', zorder=4)	

		self.fig.canvas.draw()
		self.fig.canvas.flush_events()


	def get_obs_dim(self):
		return 1

	def get_act_dim(self):
		return 1

if __name__ == '__main__':
	
	import time
	n_agents = 3
	env = TwoStepContEnv(n_agents = n_agents, intersection = 0.6)

	while True:	
		env.reset()
		env.render()
		done = False
		while not done:
			_, _, done, _ = env.step(np.random.uniform(size=(n_agents, 1)))
			time.sleep(0.3)
			env.render()
